from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import FaktsGrant, GrantException
from fakts.beacon import EndpointDiscovery, FaktsRetriever

try:
    from rich.prompt import Prompt
    from rich.console import Console
except ImportError as e:

    raise ImportError(
        "To use the cli, you need to install rich. Ex `pip install rich`"
    ) from e


class PrompingBeaconGrantException(GrantException):
    pass


class NoBeaconsFound(PrompingBeaconGrantException):
    pass


class CLIBeaconGrant(FaktsGrant):
    def __init__(
        self,
        *args,
        dicovery_protocol: EndpointDiscovery = None,
        retriever_protocol: FaktsRetriever = None,
        timeout=4,
        console=None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)

        self._discov = dicovery_protocol or EndpointDiscovery()
        self._retriev = retriever_protocol or FaktsRetriever()
        self._console = console or Console()
        self._timeout = timeout

    async def aload(self, previous={}, **kwargs):

        with self._console.status(
            f"Waiting {self._timeout} seconds for Beacon Answers"
        ):
            endpoints = await self._discov.ascan_list(timeout=self._timeout)

        if len(endpoints.keys()) == 0:
            raise NoBeaconsFound("We couldn't find any beacon in your local network")

        choices_name = [key for key, value in endpoints.items()]
        endpoint_name = Prompt.ask(
            "Which Endpoint do you want", choices=choices_name, default=choices_name[0]
        )

        with self._console.status(
            f"Please check your browser window to finish the setup"
        ):
            return await self._retriev.aretrieve(endpoints[endpoint_name], previous)



from fakts.grants.base import FaktsGrant
import yaml


class YamlGrant(FaktsGrant):
    def __init__(self, filepath="random.yaml") -> None:
        super().__init__()
        self.filepath = filepath

    async def aload(self, previous={}, **kwargs):
        with open(self.filepath, "r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

        return config



from typing import Type, TypeVar
from pydantic import BaseSettings
from pydantic.error_wrappers import ValidationError
import fakts
from fakts.fakts import Fakts, current_fakts


Class = TypeVar("Class")


class ConfigError(Exception):
    pass


class Config(BaseSettings):
    class Config:
        extra = "ignore"
        group = "undefined"

    @classmethod
    async def from_fakts(
        cls: Type[Class], fakts: Fakts = None, bypass_middleware=False, **overwrites
    ) -> Class:
        group = cls.__config__.group
        assert (
            group != "undefined"
        ), f"Please overwrite the Metaclass Config parameter group and point at your group {cls}"
        fakts = fakts or current_fakts.get()
        try:
            return cls(**await fakts.aget(group, bypass_middleware=bypass_middleware))
        except ValidationError as e:
            raise ConfigError(f"{fakts.fakts} was not sufficient") from e






from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import FaktsGrant
from fakts.beacon import EndpointDiscovery, FaktsRetriever



async def discover_endpoint(name_filter= None):
    discov = EndpointDiscovery()
    return await discov.ascan_first(name_filter=name_filter)




class BeaconGrant(FaktsGrant):

    def __init__(self, *args, dicovery_protocol: EndpointDiscovery = None, retriever_protocol: FaktsRetriever = None, **kwargs) -> None:

        self._discov = dicovery_protocol or EndpointDiscovery()
        self._retriev = retriever_protocol  or FaktsRetriever()

        super().__init__(*args, **kwargs)

    async def aload(self, previous={}, **kwargs):
        endpoint = await self._discov.ascan_first(**previous)
        return await self._retriev.aretrieve(endpoint, previous=previous)


class FaktsError(Exception):
    pass


class NoFaktsFound(FaktsError):
    pass


class NoGrantConfigured(FaktsError):
    pass


class GroupsNotFound(FaktsError):
    pass


class NoGrantSucessfull(FaktsError):
    pass



import asyncio
from fakts.beacon import EndpointBeacon, FaktsEndpoint
from rich.prompt import Prompt
import argparse




def main(name=None, url=None):
    if not name:
        name = Prompt.ask("How do you want this beacon to be advertisted as?", default="Arkitekt")

    if not url:
        url = Prompt.ask("Which Setup Uri do you want to advertise?", default="http://localhost:3000/setupapp")

    beacon = EndpointBeacon(advertised_endpoints= [FaktsEndpoint(url=url, name=name)])
    asyncio.run(beacon.run())


def entrypoint():
    parser = argparse.ArgumentParser(description = 'Say hello')
    parser.add_argument('--url', type=str, help='The Name of this script')
    parser.add_argument('--name', type=bool, help='Do you want to refresh')
    args = parser.parse_args()

    main(name=args.name, url=args.url)


if __name__ == "__main__":
    entrypoint()


from abc import abstractmethod
from typing import Generic, Type, TypeVar
from pydantic.main import BaseModel
from fakts.beacon.beacon import FaktsEndpoint
import webbrowser
import asyncio
import uuid
from aiohttp import web
from urllib.parse import quote, urlencode, parse_qs
from koil import koil
import json
import logging

logger = logging.getLogger(__name__)


class RetrieverException(Exception):
    pass


class IncorrectStateException(Exception):
    pass


def wrapped_get_future(future: asyncio.Future, state: str):
    async def web_token_response(request):
        loop = asyncio.get_event_loop()
        logger.info("Received Reply from user")
        result = parse_qs(request.query_string)
        qs_state = result["state"][0]
        config = json.loads(result["config"][0])

        if qs_state != state:
            loop.call_soon_threadsafe(
                future.set_exception, RetrieverException("Danger! Invalid State")
            )
            return web.Response(text="Error! Invalid State.")

        loop.call_soon_threadsafe(future.set_result, config)
        return web.Response(text="You can close me now !")

    return web_token_response


async def wait_for_get(
    starturl,
    previous={},
    redirect_host="localhost",
    redirect_port=6767,
    redirect_path="/",
    timeout=400,
    print_function=False,
    handle_signals=False,
):

    state = str(uuid.uuid4())

    params = {
        "state": state,
        "redirect_uri": f"http://{redirect_host}:{redirect_port}{redirect_path}",
        "scopes": previous.get("herre", {}).get("scopes", None),
        "name": previous.get("herre", {}).get("name", None),
        "client_id": previous.get("herre", {}).get("client_id", None),
    }

    querystring = urlencode(
        {key: value for key, value in params.items() if value != None}
    )

    webbrowser.open_new(starturl + "?" + querystring)

    token_future = asyncio.get_event_loop().create_future()

    app = web.Application()

    app.router.add_get(redirect_path, wrapped_get_future(token_future, state))

    webserver_task = asyncio.get_event_loop().create_task(
        web._run_app(
            app,
            host=redirect_host,
            port=redirect_port,
            print=print_function,
            handle_signals=handle_signals,
        )
    )
    done, pending = await asyncio.wait(
        [token_future, webserver_task, asyncio.sleep(timeout)],
        return_when=asyncio.FIRST_COMPLETED,
    )

    for tf in done:
        if tf == token_future:
            post_json = tf.result()
        else:
            post_json = None

    for task in pending:
        task.cancel()

        try:
            await task
        except asyncio.CancelledError:
            pass

    if not post_json:
        raise RetrieverException("No Post Data Received")
    return post_json


class FaktsRetriever:
    REDIRECT_HOST = "localhost"
    REDIRECT_PORT = 6767
    REDIRECT_PATH = "/"

    def __init__(
        self,
        *args,
        redirect_host=None,
        redirect_port=None,
        redirect_path=None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.redirect_host = redirect_host or self.REDIRECT_HOST
        self.redirect_port = redirect_port or self.REDIRECT_PORT
        self.redirect_path = redirect_path or self.REDIRECT_PATH

    async def aretrieve(self, config: FaktsEndpoint, previous={}):
        post_data = await wait_for_get(
            config.url,
            previous=previous,
            redirect_host=self.redirect_host,
            redirect_port=self.redirect_port,
            redirect_path=self.redirect_path,
        )
        return post_data

    def retrieve(self, config, as_task=True, **kwargs):
        return koil(self.aretrieve(config, **kwargs), as_task=as_task)



from fakts.grants.base import GrantException, FaktsGrant
import yaml
from koil.qt import QtCoro
from qtpy import QtWidgets


class NoFileSelected(GrantException):
    pass


class QtSelectYaml(QtWidgets.QFileDialog):
    @classmethod
    def ask(cls, text, parent=None):
        filepath, weird = cls.getOpenFileName(parent=parent, caption=text)
        return filepath


class QtYamlGrant(FaktsGrant, QtWidgets.QWidget):
    def __init__(self) -> None:
        super().__init__()
        self.get_file_coro = QtCoro(self.on_open_filepath, autoresolve=True)

    def on_open_filepath(self, ref):
        filepath = QtSelectYaml.ask("Select a Yaml", parent=self)
        return filepath

    async def aload(self, previous={}, **kwargs):
        print("Here?")
        filepath = await self.get_file_coro.acall()
        print(filepath)
        with open(filepath, "r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

        print("Config")
        print(config)
        return config



import asyncio
import contextvars
from typing import List, Type
from fakts.errors import (
    GroupsNotFound,
    NoFaktsFound,
    NoGrantConfigured,
    NoGrantSucessfull,
)
from fakts.middleware.base import FaktsMiddleware
from fakts.utils import update_nested
from koil import koil
import yaml
from fakts.grants.base import FaktsGrant
import os
from fakts.grants.yaml import YamlGrant
from fakts.middleware.environment.overwritten import OverwrittenEnvMiddleware
import logging
import sys

from koil.decorators import koilable
from koil.helpers import unkoil

logger = logging.getLogger(__name__)
current_fakts = contextvars.ContextVar("current_fakts")


@koilable(add_connectors=True)
class Fakts:
    def __init__(
        self,
        *args,
        grants=[],
        middlewares=[],
        assert_groups=[],
        fakts_path="fakts.yaml",
        subapp: str = None,
        hard_fakts={},
        force_refresh=False,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.grants: List[FaktsGrant] = grants
        self.middlewares: List[FaktsMiddleware] = middlewares
        self.hard_fakts = hard_fakts
        self.fakts = {}
        self.assert_groups = set(assert_groups)
        self.subapp = subapp
        self.fakts_path = f"{subapp}.{fakts_path}" if subapp else fakts_path
        self._lock = None
        self.force_refresh = force_refresh
        self.loaded = False

        try:
            with open(self.fakts_path, "r") as file:
                config = yaml.load(file, Loader=yaml.FullLoader)
                self.fakts = update_nested(self.hard_fakts, config)
        except Exception as e:
            print(e)
            logger.info("No fakts file found. We will have to use grants!")

        if self.force_refresh:
            self.fakts = {}

        if not self.fakts:
            assert (
                len(self.grants) >= 1
            ), f"No grants configured and we did not find fakts at path {self.fakts_path}. Please make sure you configure fakts correctly."

        for grant in self.grants:
            assert hasattr(grant, "aload"), "Grant must have an aload method"
            assert not isinstance(grant, type), "SHould not be a class"

        print(f"Fakts are {self.fakts}")

    async def aget(self, group_name: str, bypass_middleware=False, auto_load=True):
        """Get Config

        Gets the currently active configuration for the group_name. This is a loop
        save function, and will guard the current fakts state through an async lock.

        Steps:
            1. Acquire lock.
            2. If not yet loaded and auto_load is True, load (reloading should be done seperatily)
            3. Pass through middleware (can be opt out by setting bypass_iddleware to True)
            4. Return groups fakts

        Args:
            group_name (str): The group name in the fakts
            bypass_middleware (bool, optional): Bypasses the Middleware (e.g. no overwrites). Defaults to False.
            auto_load (bool, optional): Should we autoload the configuration through grants if nothing has been set? Defaults to True.

        Returns:
            dict: The active fakts
        """

        assert (
            self.loaded
        ), "Fakts not loaded, please load first. (By entering the fakts context)"

        config = {**self.fakts}

        if not bypass_middleware:
            for middleware in self.middlewares:
                additional_config = await middleware.aparse(previous=config)
                config = update_nested(config, additional_config)

        for subgroup in group_name.split("."):
            try:
                config = config[subgroup]
            except KeyError as e:
                logger.error(f"Could't find {subgroup} in {config}")
                config = {}

        return config

    async def arefresh(self):
        await self.aload()

    def get(self, *args, **kwargs):
        return unkoil(self.aget, *args, **kwargs)

    @property
    def healthy(self):
        return self.assert_groups.issubset(set(self.fakts.keys()))

    async def aload(self):
        if not self.force_refresh:
            if self.healthy:
                self.loaded = True
                return

        if len(self.grants) == 0:
            raise NoGrantConfigured(
                "Local fakts were insufficient and fakts has no grants configured. Please add a grant to your fakts instance or initialize your fakts instance with a Grant"
            )

        grant_exceptions = {}
        for grant in self.grants:
            try:
                additional_fakts = await grant.aload(previous=self.fakts)
                self.fakts = update_nested(self.fakts, additional_fakts)
            except Exception as e:
                grant_exceptions[grant.__class__.__name__] = e

        print(grant_exceptions)
        if not self.assert_groups.issubset(set(self.fakts.keys())):

            error_description = (
                f"This might be due to following exceptions in grants {grant_exceptions}"
                if grant_exceptions
                else "All Grants were sucessful. But none retrieved these keys!"
            )

            raise GroupsNotFound(
                f"Could not find {self.assert_groups - set(self.fakts.keys())}. "
                + error_description
            )

        if not self.fakts:
            raise NoGrantSucessfull(f"No Grants sucessfull {grant_exceptions}")

        if self.fakts_path:
            with open(self.fakts_path, "w") as file:
                yaml.dump(self.fakts, file)

        self.loaded = True

    async def adelete(self):
        self.fakts = {}

        if self.fakts_path:
            os.remove(self.fakts_path)

    def load(self, **kwargs):
        return unkoil(self.aload, **kwargs)

    def delete(self, **kwargs):
        return unkoil(self.adelete, **kwargs)

    async def __aenter__(self):
        self._token = current_fakts.set(self)
        await self.aload()
        return self

    async def __aexit__(self, *args, **kwargs):
        current_fakts.reset(self._token)



from koil import koil


class MiddlewareException(Exception):
    pass


class FaktsMiddleware:


    async def aparse(self, previous = {}, **kwargs):
        raise NotImplementedError()

    def parse(self, previous ={}, as_task=False, **kwargs):
        return koil(self.aparse(previous=previous, **kwargs), as_task=as_task)



from koil import koil


class GrantException(Exception):
    pass


class FaktsGrant:


    async def aload(self, previous = {}, **kwargs):
        raise NotImplementedError()

    def load(self, previous ={}, as_task=False, **kwargs):
        return koil(self.load(previous=previous, **kwargs), as_task=as_task)



import os

from fakts.middleware.base import FaktsMiddleware
import logging


logger = logging.getLogger(__name__)


def parse_env_value_to_python(value: str):
    if value in ("True", "true"):
        return True
    if value in ("false", "False"):
        return False
    return value


class OverwrittenEnvMiddleware(FaktsMiddleware):
    """Takes the previous Grants input
    and checks for environmental overwrites


    Args:
        FaktsGrant ([type]): [description]
    """

    def __init__(self, prepend=None, delimiter="_") -> None:
        super().__init__()
        self.prepend = prepend
        self.delimiter = delimiter

    async def aparse(self, previous={}, **kwargs):

        updated = {}
        try:
            for key, value in os.environ.items():
                if self.prepend:
                    if not key.startswith(self.prepend):
                        continue
                    key = key[len(self.prepend) :]

                path = list(map(lambda x: x.lower(), key.split(self.delimiter)))
                if path[0] in previous:
                    ref = updated
                    for part in path[:-1]:
                        if part not in ref:
                            ref[part] = {}
                        ref = ref[part]
                    ref[path[-1]] = value

        except Exception as e:
            logger.exception(e)

        return updated



from typing import List
from fakts.grants.base import FaktsGrant
from fakts.beacon import EndpointDiscovery, FaktsRetriever
import asyncio
from functools import reduce

from fakts.utils import update_nested


class ParallelGrant(FaktsGrant):

    def __init__(self, parellized_grants: List[FaktsGrant], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.parellized_grants = parellized_grants


    async def aload(self, previous = {}, **kwargs):

        config_futures = [grant.aload(**kwargs) for grant in self.parellized_grants]
        configs = await asyncio.gather(config_futures)
        return reduce(lambda x, y : update_nested(x, y), configs,  previous)


from qtpy.QtCore import Signal
from fakts.grants.beacon import BeaconGrant
from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import GrantException
from qtpy import QtWidgets
import asyncio
import logging
from koil.qt import QtCoro, QtFuture

logger = logging.getLogger(__name__)


class RetrieveDialog(QtWidgets.QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle("Check your Broswer")
        self.button = QtWidgets.QPushButton("Cancel")
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.button)
        self.setLayout(self.layout)


class SelfScanWidget(QtWidgets.QWidget):
    user_endpoint = Signal(FaktsEndpoint)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout()
        self.lineEdit = QtWidgets.QLineEdit()
        self.addButton = QtWidgets.QPushButton("Scan")

        self.layout.addWidget(self.lineEdit)
        self.layout.addWidget(self.addButton)
        self.addButton.clicked.connect(self.on_add)
        self.setLayout(self.layout)

    def on_add(self):
        host = self.lineEdit.text()
        url = f"http://{host}/setupapp"
        endpoint = FaktsEndpoint(url=url, name="Self Added")
        self.user_endpoint.emit(endpoint)


class UserCancelledException(GrantException):
    pass


class QtSelectableBeaconGrant(BeaconGrant, QtWidgets.QDialog):
    new_endpoint = Signal(FaktsEndpoint)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWindowTitle("Search Endpoints...")

        self.retrieve_dialog = RetrieveDialog(parent=self)

        self.show_coro = QtCoro(lambda f: self.show(), autoresolve=True)
        self.hide_coro = QtCoro(lambda f: self.hide(), autoresolve=True)

        self.select_endpoint = QtCoro(self.demand_selection_of_endpoint)
        self.select_endpoint_future = None

        self.new_endpoint.connect(self.on_new_endpoint)

        self.endpoints = []

        self.listWidget = QtWidgets.QListWidget()

        self.scanWidget = SelfScanWidget()

        QBtn = QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.rejected.connect(self.on_reject)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.listWidget)
        self.layout.addWidget(self.scanWidget)
        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def demand_selection_of_endpoint(self, future: QtFuture):
        self.select_endpoint_future = future

    def on_endpoint_clicked(self, item):
        index = self.listWidget.indexFromItem(item).row()
        self.select_endpoint_future.resolve(self.endpoints[index])

    def on_reject(self):
        if self.select_endpoint_future:
            self.select_endpoint_future.reject(
                Exception("User cancelled the this Grant without selecting a Beacon")
            )
        self.reject()

    def closeEvent(self, event):
        # do stuff
        if self.select_endpoint_future:
            self.select_endpoint_future.reject(
                Exception("User cancelled the this Grant without selecting a Beacon")
            )

        event.accept()  # let the window close

    def on_new_endpoint(self, config):
        self.listWidget.clear()

        self.endpoints.append(config)

        for endpoint in self.endpoints:
            self.listWidget.addItem(f"{endpoint.name} at {endpoint.url}")

        self.listWidget.itemClicked.connect(self.on_endpoint_clicked)

    async def emit_endpoints(self):
        try:
            async for endpoint in self._discov.ascan_gen():
                self.new_endpoint.emit(endpoint)
        except Exception as e:
            print(e)

    async def aload(self, previous={}, **kwargs):
        loop = asyncio.get_event_loop()
        emitting_task = loop.create_task(self.emit_endpoints())
        try:

            await self.show_coro.acall()
            try:
                endpoint = await self.select_endpoint.acall()

                konfik = await self._retriev.aretrieve(endpoint, previous=previous)

                await self.hide_coro.acall()

            finally:
                emitting_task.cancel()
                try:
                    await emitting_task
                except asyncio.CancelledError as e:
                    logger.info("Cancelled the Discovery task")
        except Exception as e:
            logger.exception(e)
            raise e

        return konfik



from fakts import Fakts
from qtpy import QtCore, QtWidgets
from koil.qt import QtFuture, QtCoro


class QtFakts(QtWidgets.QWidget, Fakts):
    loaded_signal = QtCore.Signal()
    deleted_signal = QtCore.Signal()
    error_signal = QtCore.Signal(Exception)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QVBoxLayout()
        self.setWindowTitle("Retrieval Wizard")

        self.show_coro = QtCoro(self.handle_show)
        self.hide_coro = QtCoro(self.handle_hide)

        self.title = QtWidgets.QLabel("Konfig Wizard")
        self.title.setAlignment(QtCore.Qt.AlignCenter)
        self.layout.addWidget(self.title)
        self.subtitle = QtWidgets.QLabel("These are the grants we are able to use")
        self.layout.addWidget(self.subtitle)

        for grant in self.grants:
            grant_title = QtWidgets.QLabel(grant.__class__.__name__)
            self.layout.addWidget(grant_title)

        QBtn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        self.buttonBox = QtWidgets.QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.on_start)
        self.buttonBox.rejected.connect(self.on_reject)

        self.layout.addWidget(self.buttonBox)
        self.setLayout(self.layout)

    def handle_show(self, future: QtFuture):
        self.show()
        self.opening_future = future

    def handle_hide(self, future: QtFuture):
        self.close()
        future.resolve()

    def on_start(self):
        self.opening_future.resolve()
        self.accept()

    def on_reject(self):
        self.opening_future.reject(Exception("User denied login"))
        self.reject()

    async def aload(self, *args, **kwargs):
        await self.show_coro.acall()  # await user starts
        try:
            nana = await super().aload()
            self.loaded_signal.emit()
            await self.hide_coro.acall()
            return nana
        except Exception as e:
            self.error_signal.emit(e)
            await self.hide_coro.acall()
            raise e

    async def adelete(self):
        nana = await super().adelete()
        self.deleted_signal.emit()
        return nana



from typing import List
from pydantic import BaseModel
from socket import socket, AF_INET, SOCK_DGRAM, SOL_SOCKET, SO_BROADCAST
import asyncio
import json
from koil import unkoil
import logging


logger = logging.getLogger(__name__)


class FaktsEndpoint(BaseModel):
    url: str = "http://localhost:3000/setupapp"
    name: str = "default"


class BeaconProtocol(asyncio.DatagramProtocol):
    pass


class EndpointBeacon:
    BIND = ""
    BROADCAST_PORT = 45678
    BROADCAST_ADDRESS = "<broadcast>"
    MAGIC_PHRASE = "beacon-fakts"

    def __init__(
        self,
        broadcast_port=None,
        bind=None,
        magic_phrase=None,
        broadcast_adress=None,
        advertised_endpoints: List[FaktsEndpoint] = [],
        interval=1,
    ) -> None:
        self.broadcast_port = broadcast_port or self.BROADCAST_PORT
        self.bind = bind or self.BIND
        self.magic_phrase = magic_phrase or self.MAGIC_PHRASE
        self.broadcast_adress = broadcast_adress or self.BROADCAST_ADDRESS

        self.advertised_endpoints = advertised_endpoints
        assert len(self.advertised_endpoints) > 0, "No config points provided"
        self.interval = interval

    def endpoint_to_message(self, config: FaktsEndpoint):
        return bytes(self.magic_phrase + json.dumps(config.dict()), "utf8")

    async def arun(self):

        s = socket(AF_INET, SOCK_DGRAM)  # create UDP socket
        s.bind((self.bind, 0))
        s.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)  # this is a broadcast socket

        loop = asyncio.get_event_loop()
        transport, pr = await loop.create_datagram_endpoint(BeaconProtocol, sock=s)

        messages = [
            self.endpoint_to_message(config) for config in self.advertised_endpoints
        ]

        while True:
            for message in messages:
                transport.sendto(message, (self.broadcast_adress, self.broadcast_port))
                logger.info(f"Send Message {message}")

            await asyncio.sleep(self.interval)

    def run(self):
        return unkoil(self.arun)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fakts',
 'fakts.beacon',
 'fakts.cli',
 'fakts.config',
 'fakts.grants',
 'fakts.grants.cli',
 'fakts.grants.meta',
 'fakts.grants.qt',
 'fakts.middleware',
 'fakts.middleware.environment']

package_data = \
{'': ['*'], 'fakts.grants.qt': ['assets/*']}

install_requires = \
['PyYAML>=5.2',
 'QtPy>=1.11.2,<2.0.0',
 'koil>=0.1.83,<0.2.0',
 'pydantic>=1.8.2,<2.0.0']

setup_kwargs = {
    'name': 'fakts',
    'version': '0.1.42',
    'description': 'asynchronous configuration protocol for dynamic client-server relations',
    'long_description': '# fakts\n\n[![codecov](https://codecov.io/gh/jhnnsrs/fakts/branch/master/graph/badge.svg?token=UGXEA2THBV)](https://codecov.io/gh/jhnnsrs/fakts)\n[![PyPI version](https://badge.fury.io/py/fakts.svg)](https://pypi.org/project/fakts/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://pypi.org/project/fakts/)\n![Maintainer](https://img.shields.io/badge/maintainer-jhnnsrs-blue)\n[![PyPI pyversions](https://img.shields.io/pypi/pyversions/fakts.svg)](https://pypi.python.org/pypi/fakts/)\n[![PyPI status](https://img.shields.io/pypi/status/fakts.svg)](https://pypi.python.org/pypi/fakts/)\n[![PyPI download day](https://img.shields.io/pypi/dm/fakts.svg)](https://pypi.python.org/pypi/fakts/)\n\n### DEVELOPMENT\n\n## Inspiration\n\nFakts tries to make the setup process between client - dynamic server relations as easy as possible.\n\n## Client - Dynamic Server\n\nIn this relation the configuration of the server is unbeknownst to the client, that means which network\naddress it can connect to retrieve its initial configuration. As both client and server need to trust\neach other this is a complex scenario to secure. Therefore fakts provided different grants to ensure different\nlevels of security.\n\n## Simple grant\n\nIn the fakts grants, fakts simply advertises itself on the network through UDP broadcasts and sends standardized configuration\nto the client. In this scenario no specific configuration on a per app basis is possible, as every client that chooses to connect to the\nserver will receive the same configuration.\n\n### Server\n\n```bash\nfakts serve fakts.yaml\n```\n\n### Client\n\n```bash\nfakts init\n```\n\nThis flow however can be secured by a password that needs to be entered, once the client wants to retrieve configuration.\n\n### Server\n\n```bash\nfakts serve fakts.yaml --password="*******"\n```\n\n## Advanced grant\n\nIn an oauth2 redirect like manner, fakts can also be used to advocate an endpoint that the app can then connect to in order to receive\nspecialised configuration through a redirect back to the client.\n\n### Beacon\n\n```bash\nfakts beacon "http://localhost:3000/beacon"\n```\n\n### Client\n\n```bash\nfakts init --client="napari"\n```\n\nIn this scenario the client will open a webbrowser with the query parameters set to the init params (in addition to a state to combat redirect attacks, and a redirect_uri) and\nwait for a redirect to its redirect_uri on localhost with the configuration in the query params\n',
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import abstractmethod
from fakts.beacon.beacon import FaktsEndpoint
from socket import socket, AF_INET, SOCK_DGRAM
import asyncio
import json
from koil import koil
import logging


logger = logging.getLogger(__name__)


class DiscoveryProtocol(asyncio.DatagramProtocol):
    pass

    def __init__(self, recvq) -> None:
        super().__init__()
        self._recvq = recvq

    def datagram_received(self, data, addr):
        self._recvq.put_nowait((data, addr))


class EndpointDiscovery:
    BROADCAST_PORT = 45678
    MAGIC_PHRASE = "beacon-fakts"
    BIND = ""

    def __init__(
        self,
        *args,
        broadcast_port=None,
        bind=None,
        magic_phrase=None,
        strict=False,
        on_new_endpoint=None,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.broadcast_port = broadcast_port or self.BROADCAST_PORT
        self.magic_phrase = magic_phrase or self.MAGIC_PHRASE
        self.bind = bind or self.BIND
        self.strict = strict
        self.discovered_endpoints = {}

    async def ascan_first(self, name_filter=None, **kwargs):

        s = socket(AF_INET, SOCK_DGRAM)  # create UDP socket
        s.bind((self.bind, self.broadcast_port))

        loop = asyncio.get_event_loop()
        read_queue = asyncio.Queue()
        transport, pr = await loop.create_datagram_endpoint(
            lambda: DiscoveryProtocol(read_queue), sock=s
        )

        while True:
            data, addr = await read_queue.get()
            try:
                data = str(data, "utf8")
                if data.startswith(self.magic_phrase):
                    endpoint = data[len(self.magic_phrase) :]

                    try:
                        endpoint = json.loads(endpoint)
                        endpoint = FaktsEndpoint(**endpoint)
                        await self.handle_new_potential_endpoint(endpoint)
                        if name_filter and endpoint.name != name_filter:
                            continue
                        return endpoint

                    except json.JSONDecodeError as e:
                        logger.error("Received Request but it was not valid json")
                        if self.strict:
                            raise e

                    except Exception as e:
                        logger.error(f"Received Request caused Exception {e}")
                        if self.strict:
                            raise e
                else:
                    logger.error(
                        f"Received Non Magic Response {data}. Maybe somebody sends"
                    )

            except UnicodeEncodeError as e:
                logger.info("Couldn't decode received message")
                if self.strict:
                    raise e

    async def ascan_gen(self, name_filter=None, **kwargs):

        s = socket(AF_INET, SOCK_DGRAM)  # create UDP socket
        s.bind((self.bind, self.broadcast_port))

        loop = asyncio.get_event_loop()
        read_queue = asyncio.Queue()
        transport, pr = await loop.create_datagram_endpoint(
            lambda: DiscoveryProtocol(read_queue), sock=s
        )

        while True:
            data, addr = await read_queue.get()
            try:
                data = str(data, "utf8")
                if data.startswith(self.magic_phrase):
                    endpoint = data[len(self.magic_phrase) :]

                    try:
                        endpoint = json.loads(endpoint)
                        endpoint = FaktsEndpoint(**endpoint)
                        if name_filter and endpoint.name != name_filter:
                            continue
                        if endpoint.name not in self.discovered_endpoints:
                            yield endpoint
                            self.discovered_endpoints[endpoint.name] = endpoint

                    except json.JSONDecodeError as e:
                        logger.error("Received Request but it was not valid json")
                        if self.strict:
                            raise e

                    except Exception as e:
                        logger.error(f"Received Request caused Exception {e}")
                        if self.strict:
                            raise e
                else:
                    logger.error(
                        f"Received Non Magic Response {data}. Maybe somebody sends"
                    )

            except UnicodeEncodeError as e:
                logger.error("Couldn't decode received message")
                if self.strict:
                    raise e

    async def ascan_list(self, timeout=5, **kwargs):
        async def appender():
            async for i in self.ascan_gen(**kwargs):
                logger.debug(f"Disocvered {i} in Time")

        try:
            appending_task = await asyncio.wait_for(
                appender(), timeout=timeout
            )  # async for should cancel this task
        except asyncio.TimeoutError as e:
            logger.info("Stopped checking")
            return self.discovered_endpoints

    def scan(self, **kwargs):
        return koil(self.ascan(**kwargs))



import collections.abc

def update_nested(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_nested(d.get(k, {}), v)
        else:
            d[k] = v
    return d


from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import FaktsGrant
from fakts.beacon import EndpointDiscovery, FaktsRetriever


class EndpointGrant(FaktsGrant):
    def __init__(
        self, endpoint: FaktsEndpoint = None, retrieve_protocol: FaktsRetriever = None
    ) -> None:
        self._endpoint = endpoint or FaktsEndpoint()
        self._retriev = retrieve_protocol or FaktsRetriever()

        super().__init__()

    async def aload(self, previous={}, **kwargs):
        return await self._retriev.aretrieve(self._endpoint, previous=previous)
